# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.release_flow._cd_release_id import (
    ReleaseIdentityError,
    identify_release_id,
)


class TestIdentifyReleaseId:
    def test_clean_feature(self):
        expected_id = "0.0.0-post.000"

        result = identify_release_id("refs/heads/feature/some/path")
        assert result == expected_id

        # Assume PR is only valid for feature/*->main
        result = identify_release_id("refs/pull/1/merge")
        assert result == expected_id

    def test_clean_development(self):
        expected_id = "0.0.0-post.000"

        result = identify_release_id("refs/heads/main")
        assert result == expected_id

    def test_clean_qa(self):
        expected_id = "3.14.159-alpha.13"

        result = identify_release_id("refs/tags/3.14.159-alpha.13")
        assert result == expected_id

    def test_clean_staging(self):
        expected_id = "3.14.159-beta.13"

        result = identify_release_id("refs/tags/3.14.159-beta.13")
        assert result == expected_id

    def test_clean_production(self):
        expected_id = "3.14.159"

        result = identify_release_id("refs/tags/3.14.159")
        assert result == expected_id

    def test_dirty_qa(self):
        """Only strictly semantic version alpha tags allowed."""
        with pytest.raises(
            ReleaseIdentityError,
            match="^Unable to match git reference to any release id",
        ):
            identify_release_id("refs/tags/3.14.159-alpha")

        with pytest.raises(
            ReleaseIdentityError,
            match="^Unable to match git reference to any release id",
        ):
            identify_release_id("refs/tags/3.14.159-alpha13")

    def test_dirty_staging(self):
        """Only strictly semantic version beta tags allowed."""
        with pytest.raises(
            ReleaseIdentityError,
            match="^Unable to match git reference to any release id",
        ):
            identify_release_id("refs/tags/3.14.159-beta")

        with pytest.raises(
            ReleaseIdentityError,
            match="^Unable to match git reference to any release id",
        ):
            identify_release_id("refs/tags/3.14.159-beta13")

    def test_dirty_production(self):
        """Only strictly semantic releases are allowed."""
        with pytest.raises(
            ReleaseIdentityError,
            match="^Unable to match git reference to any release id",
        ):
            identify_release_id("refs/tags/3.14.159.54")

    def test_unknown_state(self, mocker):
        mocker.patch(
            "foodx_devops_tools.release_flow._cd_release_id"
            ".identify_release_state"
        )
        with pytest.raises(
            ReleaseIdentityError,
            match="^Unable to match git reference to any release id",
        ):
            identify_release_id("refs/tags/3.14.159.54")



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools._to import StructuredTo, StructuredToError


def test_default():
    under_test = StructuredTo()

    assert under_test.frame is None
    assert under_test.application is None
    assert under_test.step is None


class TestFromSpecifier:
    def test_all_clean(self):
        under_test = StructuredTo.from_specifier(
            "this_frame.this_application.this_step"
        )

        assert under_test.frame == "this_frame"
        assert under_test.application == "this_application"
        assert under_test.step == "this_step"

    def test_missing_application_clean(self):
        under_test = StructuredTo.from_specifier("this_frame")

        assert under_test.frame == "this_frame"
        assert under_test.application is None
        assert under_test.step is None

    def test_missing_range_clean(self):
        under_test = StructuredTo.from_specifier("this_frame.this_application")

        assert under_test.frame == "this_frame"
        assert under_test.application == "this_application"
        assert under_test.step is None

    def test_none_specifier_default(self):
        under_test = StructuredTo.from_specifier(None)

        assert under_test.frame is None
        assert under_test.application is None
        assert under_test.step is None

    def test_bad_specifier_raises(self):
        with pytest.raises(
            StructuredToError, match=r"^Bad structured deployment specifier"
        ):
            StructuredTo.from_specifier("bad&specifier")



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib


def initialize_filesystem(file_name: pathlib.Path, file_content: str):
    with file_name.open("w") as f:
        f.write(file_content)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Pipeline release flow coordination."""

from ._cd_release_id import identify_release_id  # noqa: F401
from ._cd_release_state import identify_release_state  # noqa: F401
from ._exceptions import ReleaseStateError  # noqa: F401
from ._group import release_flow  # noqa: F401



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.deploy_me._deployment import (
    DeploymentState,
    assess_results,
)


class TestAssessResults:
    @pytest.mark.asyncio
    async def test_success(self):
        mock_results = [
            DeploymentState(code=DeploymentState.ResultType.success),
            DeploymentState(code=DeploymentState.ResultType.success),
        ]

        result = await assess_results(mock_results)

        assert result.code == DeploymentState.ResultType.success

    @pytest.mark.asyncio
    async def test_fail(self):
        mock_results = [
            DeploymentState(code=DeploymentState.ResultType.success),
            DeploymentState(code=DeploymentState.ResultType.failed),
        ]

        result = await assess_results(mock_results)

        assert result.code == DeploymentState.ResultType.failed



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Exception declarations for package internal consumption."""


class ConfigurationPathsError(Exception):
    """Problem with configuration paths, or acquiring configuration."""


class DeploymentsDefinitionError(Exception):
    """Problem loading deployment definitions."""


class FrameDefinitionsError(Exception):
    """Problem loading frame definitions."""


class FolderPathError(Exception):
    """Problem acquiring frame folder path."""


class PipelineConfigurationError(Exception):
    """Problem occurred loading pipeline configuration."""


class PuffMapDefinitionsError(Exception):
    """Problem parsing puff map definitions."""


class ReleaseStatesDefinitionError(Exception):
    """Problem loading release state definitions."""


class ServicePrincipalsError(Exception):
    """Problem loading service principal secrets."""


class StaticSecretsError(Exception):
    """Problem loading static secrets."""


class SubscriptionsDefinitionError(Exception):
    """Problem loading subscription definitions."""


class SystemsDefinitionError(Exception):
    """Problem loading system definitions."""


class TenantsDefinitionError(Exception):
    """Problem loading tenant definitions."""


class PipelineViewError(Exception):
    """Problem with view data."""


class ClientsDefinitionError(Exception):
    """Problem loading client definitions."""


class TemplateContextError(Exception):
    """Problem loading template context variable definitions."""



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import os
import pathlib
import tempfile
import typing

from foodx_devops_tools.pipeline_config import load_static_secrets
from foodx_devops_tools.utilities.ansible import _encrypt_vault


@contextlib.contextmanager
def encrypted_files(
    content: typing.Dict[str, str], password: str
) -> typing.Generator[typing.List[pathlib.Path], None, None]:
    with tempfile.TemporaryDirectory() as dir:
        dir_path = pathlib.Path(dir)
        encrypted_file_paths = set()
        for file_name, file_content in content.items():
            this_file = dir_path / file_name

            password_file_path = dir_path / "password-file"
            with password_file_path.open("w") as f:
                f.write(password)

            unencrypted_file_path = dir_path / "plain.file"
            with unencrypted_file_path.open("w") as f:
                f.write(file_content)

            _encrypt_vault(this_file, password_file_path, unencrypted_file_path)

            os.remove(unencrypted_file_path)
            os.remove(password_file_path)

            assert this_file.is_file()
            with this_file.open("r") as f:
                encrypted_content = f.read()

            assert encrypted_content.startswith("$ANSIBLE_VAULT;1.1;AES256")

            encrypted_file_paths.add(this_file)

        yield encrypted_file_paths


def test_load():
    file_text = {
        "f1": """---
static_secrets:
  s1:
    s1k1: s1k1v
    s1k2: s1k2v
  s2:
    s2k1: s2k1v
""",
        "f2": """---
static_secrets:
  s3:
    s3k1: s3k1v
""",
    }
    decrypt_token = "somesecret"
    with encrypted_files(file_text, decrypt_token) as encrypted_dir_paths:
        result = load_static_secrets(encrypted_dir_paths, decrypt_token)

    assert len(result.static_secrets) == 3
    assert result.static_secrets == {
        "s1": {
            "s1k1": "s1k1v",
            "s1k2": "s1k2v",
        },
        "s2": {"s2k1": "s2k1v"},
        "s3": {"s3k1": "s3k1v"},
    }



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Run authentication tests against an actual Azure Cloud instance"""

import json
import logging

import pytest

from foodx_devops_tools.azure.cloud import (
    AzureCredentials,
    login_service_principal,
)
from foodx_devops_tools.azure.cloud.exceptions import AzureAuthenticationError

log = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_clean(service_principal_credentials):
    await login_service_principal(service_principal_credentials)


@pytest.mark.asyncio
async def test_failed():
    MOCK_CREDENTIALS = AzureCredentials(
        name="mock_user",
        userid="123abc",
        secret="verysecret",
        subscription="this_subscription",
        tenant="this_tenant",
    )
    with pytest.raises(AzureAuthenticationError):
        await login_service_principal(MOCK_CREDENTIALS)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.release_flow_entry import release_flow
from tests.ci.support.click_runner import click_runner  # noqa: F401


def test_help(click_runner):
    mock_input = ["--help"]

    result = click_runner.invoke(release_flow, mock_input)

    assert result.exit_code == 0
    assert "Release flow command group." in result.output



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy

import pytest

from foodx_devops_tools.pipeline_config.exceptions import PipelineViewError
from foodx_devops_tools.pipeline_config.views import (
    DeploymentContext,
    DeploymentTuple,
    DeploymentView,
    ReleaseView,
)
from tests.ci.support.pipeline_config import MOCK_CONTEXT, MOCK_RESULTS


@pytest.fixture()
def mock_context() -> DeploymentContext:
    this_context = copy.deepcopy(MOCK_CONTEXT)
    this_context.release_state = "r2"

    return this_context


@pytest.fixture()
def this_deployment_view(mock_pipeline_config):
    def _apply(mock_results: dict, this_context: DeploymentContext):
        expected_state = DeploymentTuple(
            client="c1", release_state="r1", system="sys1"
        )
        release_view = ReleaseView(
            mock_pipeline_config(mock_results), this_context
        )
        under_test = DeploymentView(release_view, expected_state)

        return under_test

    return _apply


class TestDeploymentView:
    def test_clean(self, mock_context, mock_pipeline_config):
        expected_state = DeploymentTuple(
            client="c1", release_state="r1", system="sys1"
        )
        release_view = ReleaseView(mock_pipeline_config(), mock_context)
        under_test = DeploymentView(release_view, expected_state)

        assert under_test.release_view == release_view
        assert under_test.deployment_tuple == expected_state

    def test_bad_deployment_state_raises(
        self, mock_context, mock_pipeline_config
    ):
        release_view = ReleaseView(mock_pipeline_config(), mock_context)
        with pytest.raises(PipelineViewError, match=r"^Bad client"):
            DeploymentView(
                release_view,
                DeploymentTuple(
                    client="bad_client", release_state="r2", system="sys2"
                ),
            )
        with pytest.raises(PipelineViewError, match=r"^Bad release state"):
            DeploymentView(
                release_view,
                DeploymentTuple(
                    client="c1", release_state="bad_release", system="sys2"
                ),
            )
        with pytest.raises(PipelineViewError, match=r"^Bad system"):
            DeploymentView(
                release_view,
                DeploymentTuple(
                    client="c1", release_state="r2", system="bad_system"
                ),
            )

    def test_subscriptions_clean(self, mock_context, mock_pipeline_config):
        expected_state = DeploymentTuple(
            client="c1", release_state="r1", system="sys1"
        )
        release_view = ReleaseView(mock_pipeline_config(), mock_context)
        under_test = DeploymentView(release_view, expected_state)

        assert under_test.release_view == release_view
        assert under_test.deployment_tuple == expected_state

        assert under_test.subscriptions

    def test_subscriptions_empty(self, mock_context, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        under_test = DeploymentView(
            release_view,
            DeploymentTuple(client="c2", release_state="r2", system="sys2"),
        )

        assert not under_test.subscriptions


@pytest.fixture()
def deployment_view_config():
    mock_results = copy.deepcopy(MOCK_RESULTS)
    mock_results["puff_map"] = {
        "frames": {
            "f1": {
                "applications": {
                    "a1": {
                        "arm_parameters_files": {
                            "r1": {
                                "sys1_c1_r1a": {
                                    "a1l1": "some/puff_map/path",
                                },
                                "sys1_c1_r1b": {
                                    "a1l1": "some/puff_map/path",
                                },
                                "sys1_c1_r1c": {
                                    "a1l1": "some/puff_map/path",
                                },
                            },
                        },
                    },
                },
            },
        },
    }
    mock_results["subscriptions"] = {
        "sys1_c1_r1a": {
            "ado_service_connection": "some-name",
            "azure_id": "abc123",
            "tenant": "t1",
        },
        "sys1_c1_r1b": {
            "ado_service_connection": "some-name",
            "azure_id": "abc123",
            "tenant": "t1",
        },
        "sys1_c1_r1c": {
            "ado_service_connection": "some-name",
            "azure_id": "abc123",
            "tenant": "t1",
        },
    }
    mock_results["service_principals"] = {
        "sys1_c1_r1a": {
            "id": "12345",
            "secret": "verysecret",
            "name": "sp_name",
        },
        "sys1_c1_r1b": {
            "id": "12345",
            "secret": "verysecret",
            "name": "sp_name",
        },
        "sys1_c1_r1c": {
            "id": "12345",
            "secret": "verysecret",
            "name": "sp_name",
        },
    }
    mock_results["static_secrets"] = {
        "sys1_c1_r1a": {"k1": "k1v1"},
        "sys1_c1_r1b": {"k1": "k1v2"},
        "sys1_c1_r1c": {"k1": "k1v2"},
    }

    return mock_results


class TestSubscriptionConstraints:
    def test_no_constraint(
        self, mock_context, deployment_view_config, this_deployment_view
    ):
        deployment_view_config["deployments"] = {
            "deployment_tuples": {
                "sys1-c1-r1": {
                    "subscriptions": {
                        "sys1_c1_r1a": {
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "this.sub",
                        },
                        "sys1_c1_r1b": {
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "other.sub",
                        },
                    },
                },
            },
            "url_endpoints": ["a", "p"],
        }
        under_test = this_deployment_view(deployment_view_config, mock_context)
        this_subscriptions = under_test.subscriptions

        assert len(under_test.subscriptions) == 2
        expected_names = {"sys1_c1_r1a", "sys1_c1_r1b"}
        assert {
            this_subscriptions[x].subscription_name
            for x in range(len(this_subscriptions))
        } == expected_names

    def test_plain_text(
        self, mock_context, deployment_view_config, this_deployment_view
    ):
        mock_context.git_ref = "b1"
        deployment_view_config["deployments"] = {
            "deployment_tuples": {
                "sys1-c1-r1": {
                    "subscriptions": {
                        "sys1_c1_r1a": {
                            "gitref_patterns": ["b1"],
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "some.where",
                        },
                        "sys1_c1_r1b": {
                            "gitref_patterns": ["b2"],
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "some.where",
                        },
                    },
                },
            },
            "url_endpoints": ["a", "p"],
        }
        under_test = this_deployment_view(deployment_view_config, mock_context)
        this_subscriptions = under_test.subscriptions

        assert len(under_test.subscriptions) == 1
        expected_names = {"sys1_c1_r1a"}
        assert {
            this_subscriptions[x].subscription_name
            for x in range(len(this_subscriptions))
        } == expected_names

    def test_regex(
        self, mock_context, deployment_view_config, this_deployment_view
    ):
        mock_context.git_ref = "b3/some/branch"
        deployment_view_config["deployments"] = {
            "deployment_tuples": {
                "sys1-c1-r1": {
                    "subscriptions": {
                        "sys1_c1_r1a": {
                            "gitref_patterns": ["^b[2-4]/.*"],
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "some.where",
                        },
                        "sys1_c1_r1b": {
                            "gitref_patterns": ["b2"],
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "some.where",
                        },
                    },
                },
            },
            "url_endpoints": ["a", "p"],
        }
        under_test = this_deployment_view(deployment_view_config, mock_context)
        this_subscriptions = under_test.subscriptions

        assert len(under_test.subscriptions) == 1
        expected_names = {"sys1_c1_r1a"}
        assert {
            this_subscriptions[x].subscription_name
            for x in range(len(this_subscriptions))
        } == expected_names

    def test_multiple(
        self, mock_context, deployment_view_config, this_deployment_view
    ):
        mock_context.git_ref = "c"
        deployment_view_config["deployments"] = {
            "deployment_tuples": {
                "sys1-c1-r1": {
                    "subscriptions": {
                        "sys1_c1_r1a": {
                            "gitref_patterns": ["^b[2-4]/.*", "c"],
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "some.where",
                        },
                        "sys1_c1_r1b": {
                            "gitref_patterns": ["c"],
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "some.where",
                        },
                        "sys1_c1_r1c": {
                            # this subscription should be excluded
                            "gitref_patterns": ["d"],
                            "locations": [{"primary": "l1"}, {"primary": "l2"}],
                            "root_fqdn": "some.where",
                        },
                    },
                },
            },
            "url_endpoints": ["a", "p"],
        }
        under_test = this_deployment_view(deployment_view_config, mock_context)
        this_subscriptions = under_test.subscriptions

        assert len(under_test.subscriptions) == 2
        expected_names = {"sys1_c1_r1a", "sys1_c1_r1b"}
        assert {
            this_subscriptions[x].subscription_name
            for x in range(len(this_subscriptions))
        } == expected_names



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with foodx_devops_tools.
#  If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest

from foodx_devops_tools.pipeline_config import StructuredName, load_frames
from foodx_devops_tools.pipeline_config._structure import FrameFile
from foodx_devops_tools.pipeline_config.exceptions import FrameDefinitionsError


@pytest.fixture
def apply_applications_test(apply_pipeline_config_test):
    def _apply(mock_content: str):
        result = apply_pipeline_config_test(mock_content, load_frames)
        return result

    return _apply


def test_single_default(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: a1_group
        a2:
          steps:
          - name: a2l1
            mode: Complete
            resource_group: a2_group
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames) == 1
    assert len(result_frames.frames["f1"].applications["a1"].steps) == 1
    assert (
        result_frames.frames["f1"].applications["a1"].steps[0].resource_group
        == "a1_group"
    )
    assert (
        not result_frames.frames["f1"]
        .applications["a1"]
        .steps[0]
        .static_secrets
    )
    assert len(result_frames.frames["f1"].applications["a2"].steps) == 1
    assert (
        result_frames.frames["f1"].applications["a2"].steps[0].resource_group
        == "a2_group"
    )
    assert result_frames.frames["f1"].folder == pathlib.Path("some/path")
    assert result_frames.frames["f1"].triggers is None
    assert (
        result_frames.frames["f1"].applications["a2"].steps[0].arm_file is None
    )
    assert (
        result_frames.frames["f1"].applications["a2"].steps[0].puff_file is None
    )
    assert result_frames.frames["f1"].applications["a2"].steps[0].name == "a2l1"
    assert result_frames.triggers is None


def test_deploy_delay(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: a1_group
          - delay_seconds: 23
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames) == 1
    assert len(result_frames.frames["f1"].applications["a1"].steps) == 2
    assert (
        result_frames.frames["f1"].applications["a1"].steps[0].resource_group
        == "a1_group"
    )
    assert (
        not result_frames.frames["f1"]
        .applications["a1"]
        .steps[0]
        .static_secrets
    )
    assert not hasattr(
        result_frames.frames["f1"].applications["a1"].steps[0], "delay_seconds"
    )
    assert (
        result_frames.frames["f1"].applications["a1"].steps[1].delay_seconds
        == 23
    )
    assert not hasattr(
        result_frames.frames["f1"].applications["a1"].steps[1], "resource_group"
    )


def test_enable_static_secrets(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: a1_group
            static_secrets: true
        a2:
          steps:
          - name: a2l1
            mode: Complete
            resource_group: a2_group
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames) == 1
    assert len(result_frames.frames["f1"].applications["a1"].steps) == 1
    assert (
        result_frames.frames["f1"].applications["a1"].steps[0].resource_group
        == "a1_group"
    )
    assert result_frames.frames["f1"].applications["a1"].steps[0].static_secrets
    assert (
        not result_frames.frames["f1"]
        .applications["a2"]
        .steps[0]
        .static_secrets
    )
    assert len(result_frames.frames["f1"].applications["a2"].steps) == 1
    assert (
        result_frames.frames["f1"].applications["a2"].steps[0].resource_group
        == "a2_group"
    )
    assert result_frames.frames["f1"].folder == pathlib.Path("some/path")
    assert result_frames.frames["f1"].triggers is None
    assert (
        result_frames.frames["f1"].applications["a2"].steps[0].arm_file is None
    )
    assert (
        result_frames.frames["f1"].applications["a2"].steps[0].puff_file is None
    )
    assert result_frames.frames["f1"].applications["a2"].steps[0].name == "a2l1"
    assert result_frames.triggers is None


def test_arm_optional(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: a1_group
            
        a2:
          steps:
          - name: a2l1
            arm_file: something.json
            mode: Complete
            resource_group: a2_group
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert result_frames.frames["f1"].applications["a2"].steps[
        0
    ].arm_file == pathlib.Path("something.json")


def test_global_path_triggers_optional(apply_applications_test):
    file_text = """---
frames:
  triggers:
    paths:
      - "some/glob/**"
      - "*/stuff/*"
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            resource_group: a1_group
            mode: Incremental
        a2:
          steps:
          - name: a2l1
            resource_group: a2_group
            mode: Complete
            arm_file: something.json
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.triggers.paths) == 2
    assert result_frames.triggers.paths[0] == "some/glob/**"
    assert result_frames.triggers.paths[1] == "*/stuff/*"


def test_frame_path_triggers_optional(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            resource_group: a1_group
            mode: Incremental
        a2:
          steps:
          - name: a2l1
            resource_group: a2_group
            mode: Complete
            arm_file: something.json
      folder: some/path
      triggers:
        paths:
          - "some/glob/**"
          - "*/stuff/*"
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames["f1"].triggers.paths) == 2
    assert result_frames.frames["f1"].triggers.paths[0] == "some/glob/**"
    assert result_frames.frames["f1"].triggers.paths[1] == "*/stuff/*"


def test_puff_optional(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            resource_group: a1_group
            mode: Incremental
        a2:
          steps:
          - name: a2l1
            resource_group: a2_group
            mode: Complete
            puff_file: something.yml
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert result_frames.frames["f1"].applications["a2"].steps[
        0
    ].puff_file == pathlib.Path("something.yml")


def test_frames_sequenced(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1:
          steps:
          - name: a1l1
            resource_group: f1a1
            mode: Complete
        a2:
          steps:
          - name: a2l1
            resource_group: f1a2
            mode: Incremental
      folder: some/f1-path
    f2:
      applications:
        a3:
          steps:
          - name: a3l1
            resource_group: f2a3
            mode: Complete
        a4:
          steps:
          - name: a4l1
            resource_group: f2a4
            mode: Incremental
      depends_on:
        - f1
      folder: some/f2-path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames) == 2
    assert "f1" in result_frames.frames
    assert "f2" in result_frames.frames
    assert result_frames.frames["f2"].depends_on[0] == "f1"


def test_applications_sequenced(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1:
          depends_on:
            - a2
          steps:
            - name: a1l1
              resource_group: f1a1
              mode: Complete
        a2:
          steps:
            - name: a2l1
              resource_group: f1a2
              mode: Incremental
      folder: some/f1-path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames["f1"].applications["a1"].depends_on) == 1
    assert result_frames.frames["f1"].applications["a1"].depends_on[0] == "a2"


def test_multiple_unsequenced(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1:
          steps:
          - name: a1l1
            resource_group: f1a1
            mode: Complete
        a2:
          steps:
          - name: a2l1
            resource_group: f1a2
            mode: Complete
      folder: some/f1-path
    f3:
      applications:
        a5:
          steps:
          - name: a5l1
            resource_group: f3a5
            mode: Complete
        a6:
          steps:
          - name: a6l1
            resource_group: f3a6
            mode: Complete
      folder: some/f3-path
    f2:
      applications:
        a3:
          steps:
          - name: a3l1
            resource_group: f2a3
            mode: Complete
        a4:
          steps:
          - name: a4l1
            resource_group: f2a4
            mode: Complete
      folder: some/f2-path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames) == 3
    assert all([x in result_frames.frames for x in ["f1", "f2", "f3"]])


def test_none_raises(apply_applications_test):
    file_text = """---
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


def test_empty_raises(apply_applications_test):
    file_text = """---
frames:
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


def test_bad_dependency_raises(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        - a1
        - a2
      folder: some/f1-path
    f2:
      applications:
        - a3
        - a4
      depends_on:
        - bad_value
      folder: some/f2-path
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


def test_missing_name_raises(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: a1_group
        a2:
          steps:
          - mode: Complete
            resource_group: a2_group
      folder: some/path
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


def test_missing_mode_raises(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: a1_group
        a2:
          steps:
          - name: a2l1
            resource_group: a2_group
      folder: some/path
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


def test_external_application_dependency_raises(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1s1
            mode: Incremental
            resource_group: a1_group
          depends_on:
            - a2
      folder: f1/path
    f2:
      applications:
        a2:
          steps:
          - name: a2s1
            mode: Incremental
            resource_group: a2_group
      folder: f2/path
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


def test_frame_application_dependency_raises(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
            - name: a1s1
              mode: Incremental
              resource_group: a1_group
          depends_on:
            # application dependency can only depend on other peer applications
            - f2
      folder: f1/path
    f2:
      applications:
        a2:
          steps:
          - name: a2s1
            mode: Incremental
            resource_group: a2_group
      folder: f2/path
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


class TestArmFilePaths:
    def test_default(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.arm_file_paths()

        assert result == {
            StructuredName(["f1", "a1", "a1l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=None,
            ),
        }

    def test_explicit(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            arm_file: arm_file.json
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.arm_file_paths()

        assert result == {
            StructuredName(["f1", "a1", "a1l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=pathlib.Path("arm_file.json"),
            )
        }

    def test_mixed(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: rgn
        a2: 
          steps:
          - name: a2l1
            mode: Incremental
            arm_file: arm_file.json
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.arm_file_paths()

        assert result == {
            StructuredName(["f1", "a1", "a1l1"]): FrameFile(
                dir=pathlib.Path("some/path/"), file=None
            ),
            StructuredName(["f1", "a2", "a2l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=pathlib.Path("arm_file.json"),
            ),
        }

    def test_subpath(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            arm_file: other/path/arm_file.json
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.arm_file_paths()

        assert result == {
            StructuredName(["f1", "a1", "a1l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=pathlib.Path("other/path/arm_file.json"),
            )
        }


class TestPuffFilePaths:
    def test_default(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.puff_file_paths()

        assert result == {
            StructuredName(["f1", "a1", "a1l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=None,
            ),
        }

    def test_explicit(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            puff_file: puff_file.yml
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.puff_file_paths()

        assert result == {
            StructuredName(["f1", "a1", "a1l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=pathlib.Path("puff_file.yml"),
            )
        }

    def test_mixed(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: rgn
        a2: 
          steps:
          - name: a2l1
            mode: Incremental
            puff_file: puff/file.yml
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.puff_file_paths()

        assert result == {
            StructuredName(["f1", "a1", "a1l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=None,
            ),
            StructuredName(["f1", "a2", "a2l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=pathlib.Path("puff/file.yml"),
            ),
        }

    def test_subpath(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            puff_file: other/path/puff_file.yml
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.puff_file_paths()

        assert result == {
            StructuredName(["f1", "a1", "a1l1"]): FrameFile(
                dir=pathlib.Path("some/path/"),
                file=pathlib.Path("other/path/puff_file.yml"),
            )
        }


class TestFrameFolders:
    def test_single(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: rgn
      folder: some/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.frame_folders()

        assert result == {
            StructuredName(["f1"]): FrameFile(
                dir=pathlib.Path("some/path"), file=pathlib.Path("")
            )
        }

    def test_multiple(self, apply_applications_test):
        file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: rgn
      folder: some/path
    f2:
      applications:
        a1: 
          steps:
          - name: a1l1
            mode: Incremental
            resource_group: rgn
      folder: f2/path
"""

        under_test = apply_applications_test(file_text)
        result = under_test.frames.frame_folders()

        assert result == {
            StructuredName(["f1"]): FrameFile(
                dir=pathlib.Path("some/path"), file=pathlib.Path("")
            ),
            StructuredName(["f2"]): FrameFile(
                dir=pathlib.Path("f2/path"), file=pathlib.Path("")
            ),
        }



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import pathlib
import tempfile
import typing

import pytest
import ruamel.yaml

from foodx_devops_tools.utilities.jinja2 import FrameTemplates

MOCK_TEMPLATE = """---
this_field: {{ this_parameter }}
"""


@contextlib.contextmanager
def mock_template_dir(
    template_content: str,
    template_name: str = "mock_template",
) -> typing.Generator[pathlib.Path, None, None]:
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)
        template_file = this_dir / template_name

        with template_file.open(mode="w") as f:
            f.write(template_content)

        yield template_file


def load_yaml(file: pathlib.Path) -> dict:
    yaml = ruamel.yaml.YAML(typ="safe")
    with file.open("r") as f:
        result = yaml.load(f)

    return result


@pytest.mark.asyncio
async def test_clean():
    parameters = {"this_parameter": "this_value"}
    with mock_template_dir(MOCK_TEMPLATE) as template_file_path:
        target_file = template_file_path.parent / "this_target"

        assert not target_file.is_file()

        under_test = FrameTemplates([template_file_path.parent])
        await under_test.apply_template(
            "mock_template", target_file, parameters
        )

        assert target_file.is_file()
        target_data = load_yaml(target_file)
        assert target_data == {"this_field": "this_value"}



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import pathlib
import tempfile
import typing

from foodx_devops_tools.pipeline_config import load_template_context


@contextlib.contextmanager
def context_files(
    content: typing.Dict[str, typing.Dict[str, str]]
) -> typing.Generator[typing.List[pathlib.Path], None, None]:
    dir_paths = set()
    file_paths = set()
    with tempfile.TemporaryDirectory() as base_path:
        for this_dir, file_data in content.items():
            dir_path = pathlib.Path(base_path) / this_dir
            dir_path.mkdir(parents=True)
            dir_paths.add(dir_path)
            for file_name, file_content in file_data.items():
                this_file = dir_path / file_name
                with this_file.open("w") as f:
                    f.write(file_content)

                file_paths.add(this_file)

        yield dir_paths, file_paths


def test_load_files():
    file_text = {
        "a": {
            "f1": """---
context:
  s1:
    s1k1: s1k1v
    s1k2: s1k2v
  s2:
    s2k1: s2k1v
""",
            "f2": """---
context:
  s3:
    s3k1: s3k1v
""",
        },
    }
    with context_files(file_text) as (dir_paths, file_paths):
        result = load_template_context(file_paths)

    assert len(result.context) == 3
    assert result.context == {
        "s1": {
            "s1k1": "s1k1v",
            "s1k2": "s1k2v",
        },
        "s2": {"s2k1": "s2k1v"},
        "s3": {"s3k1": "s3k1v"},
    }


def test_load_dirs():
    file_text = {
        "a": {
            "f1": """---
context:
  s1:
    s1k1: s1k1v
    s1k2: s1k2v
  s2:
    s2k1: s2k1v
""",
            "f2": """---
context:
  s3:
    s3k1: s3k1v
""",
        },
        "b": {
            "f1": """---
context:
  s1:
    s1k3: s1k3v
  s4:
    s4k1: s4k1v
""",
        },
    }
    with context_files(file_text) as (dir_paths, file_paths):
        result = load_template_context(file_paths)

    assert len(result.context) == 4
    assert result.context == {
        "s1": {
            "s1k1": "s1k1v",
            "s1k2": "s1k2v",
            "s1k3": "s1k3v",
        },
        "s2": {"s2k1": "s2k1v"},
        "s3": {"s3k1": "s3k1v"},
        "s4": {"s4k1": "s4k1v"},
    }


def test_deep_merge():
    """yaml object data should be merged across all levels in the data."""
    file_text = {
        "a": {
            "f1": """---
context:
  s1:
    s1k1: 
      s1k1kk1: v1
""",
            "f2": """---
context:
  s1:
    s1k1: 
      s1k1kk2: 
        s1k1kkk1: v2
""",
        },
        "b": {
            "f1": """---
context:
  s1:
    s1k1: 
      s1k1kk2: 
        s1k1kkk2: v3
""",
        },
    }
    with context_files(file_text) as (dir_paths, file_paths):
        result = load_template_context(file_paths)

    assert result.context == {
        "s1": {
            "s1k1": {
                "s1k1kk1": "v1",
                "s1k1kk2": {
                    "s1k1kkk1": "v2",
                    "s1k1kkk2": "v3",
                },
            },
        },
    }



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Template context variable deployment configuration I/O."""
import copy
import logging
import pathlib
import typing

import deepmerge  # type: ignore
import pydantic

from ._exceptions import TemplateContextError
from ._loader import load_yaml_data

log = logging.getLogger(__name__)

ENTITY_NAME = "context"

ValueType = typing.Dict[str, typing.Any]


class TemplateContext(pydantic.BaseModel):
    """Define a collection of template context variables."""

    context: ValueType


def load_template_context(
    context_paths: typing.Set[pathlib.Path],
) -> TemplateContext:
    """Load template context variables from the relevant directory."""
    try:
        context_data: dict = {
            ENTITY_NAME: dict(),
        }
        for this_path in context_paths:
            if this_path.is_file():
                log.info("loading template context, {0}".format(this_path))
                yaml_data = load_yaml_data(this_path)
                if ENTITY_NAME in yaml_data:
                    context_data = deepmerge.always_merger.merge(
                        copy.deepcopy(context_data), yaml_data
                    )
                else:
                    message = (
                        f"template context object not present in "
                        f"file, {this_path}"
                    )
                    log.error(message)
                    raise TemplateContextError(message)

        this_object = TemplateContext.parse_obj(context_data)

        return this_object
    except (NotADirectoryError, FileNotFoundError) as e:
        raise TemplateContextError(str(e)) from e



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""ARM template parameters header."""

import typing

ARMTEMPLATE_PARAMETERS_HEADER: typing.Dict[str, typing.Any] = {
    "$schema": "https://schema.management.azure.com/schemas/2019-04-01"
    "/deploymentParameters.json#",
    "contentVersion": "1.0.0.0",
    "parameters": dict(),
}



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""FoodX core deployment utility."""

from ._deployment import prepare_deployment_files  # noqa: F401
from ._main import deploy_me  # noqa: F401
from ._state import ExitState  # noqa: F401



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy

import pytest

from foodx_devops_tools.pipeline_config.exceptions import PipelineViewError
from foodx_devops_tools.pipeline_config.views import (
    DeploymentTuple,
    DeploymentView,
    IterationContext,
    ReleaseView,
    SubscriptionView,
)
from tests.ci.support.pipeline_config import MOCK_CONTEXT, MOCK_TO


class TestIterationContext:
    def test_clean(self):
        under_test = IterationContext()
        under_test.append("one")
        under_test.append("two")

        assert str(under_test) == "one.two"

    def test_empty(self):
        under_test = IterationContext()

        assert str(under_test) == ""


class TestFlattenedDeployment:
    def test_copy_add_frame(self, mock_flattened_deployment):
        expected_frame = "f1"
        under_test = mock_flattened_deployment[0]

        assert expected_frame not in under_test.data.iteration_context

        result = under_test.copy_add_frame(expected_frame)

        assert id(result) != id(under_test)
        assert expected_frame in result.data.iteration_context
        assert result.context.frame_name == expected_frame

    def test_copy_add_application(self, mock_flattened_deployment):
        expected_name = "a1"
        under_test = mock_flattened_deployment[0]

        assert expected_name not in under_test.data.iteration_context

        result = under_test.copy_add_application(expected_name)

        assert id(result) != id(under_test)
        assert expected_name in result.data.iteration_context
        assert result.context.application_name == expected_name

    def test_construct_app_fqdns(self, mock_flattened_deployment):
        under_test = mock_flattened_deployment[0]

        result = under_test.construct_app_fqdns()

        assert result == {
            "a": "a.r1a.c1.some.where",
            "p": "p.r1a.c1.some.where",
            "root": "some.where",
            "support": "support.c1.some.where",
        }

    def test_construct_app_urls(self, mock_flattened_deployment):
        under_test = mock_flattened_deployment[0]

        result = under_test.construct_app_urls()

        assert result == {
            "a": "https://a.r1a.c1.some.where",
            "p": "https://p.r1a.c1.some.where",
            "support": "https://support.c1.some.where",
        }

    def test_construct_fqdn(self, mock_flattened_deployment):
        under_test = mock_flattened_deployment[0]

        result = under_test.construct_fqdn("api")

        assert result == "api.r1a.c1.some.where"

    def test_construct_template_parameters(self, mock_flattened_deployment):
        under_test = mock_flattened_deployment[0]
        under_test.context.frame_name = "f1"
        under_test.data.template_context = {
            "v1": {"k1": 3.14},
            "v2": "vv2",
        }

        result = under_test.construct_template_parameters()

        assert result == {
            "context": {
                "v1": {"k1": 3.14},
                "v2": "vv2",
                "locations": {
                    "primary": "l1",
                    "secondary": None,
                },
                "network": {
                    "fqdns": {
                        "a": "a.r1a.c1.some.where",
                        "p": "p.r1a.c1.some.where",
                        "root": "some.where",
                        "support": "support.c1.some.where",
                    },
                    "urls": {
                        "a": "https://a.r1a.c1.some.where",
                        "p": "https://p.r1a.c1.some.where",
                        "support": "https://support.c1.some.where",
                    },
                },
                "tags": {
                    "application_name": None,
                    "client": "c1",
                    "commit_sha": "abc123",
                    "frame_name": "f1",
                    "pipeline_id": "123456",
                    "release_id": "3.1.4+local",
                    "release_state": "r1",
                    "resource_suffix": "r1a",
                    "subscription_name": "sys1_c1_r1a",
                    "system": "sys1",
                    "tenant_name": "t1",
                },
            }
        }


class TestSubscriptionView:
    def test_clean(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentTuple(client="c1", release_state="r1", system="sys1"),
        )
        SubscriptionView(
            deployment_view,
            "sys1_c1_r1a",
        )

    def test_deploy_data_clean(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentTuple(client="c1", release_state="r1", system="sys1"),
        )
        under_test = SubscriptionView(
            deployment_view,
            "sys1_c1_r1a",
        )

        result = under_test.deploy_data

        assert [x.azure_credentials.userid for x in result] == [
            "12345",
            "12345",
        ]
        assert {x.location_primary for x in result} == {"l1", "l2"}

    def test_bad_subscription_raises(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentTuple(client="c1", release_state="r1", system="sys1"),
        )
        with pytest.raises(PipelineViewError, match=r"^Bad subscription"):
            SubscriptionView(
                deployment_view,
                "bad_sub",
            )


class TestReleaseView:
    def test_clean(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        under_test = ReleaseView(mock_pipeline_config(), this_context)

        assert under_test.configuration == mock_pipeline_config()
        assert under_test.deployment_context.release_state == "r2"

    def test_bad_release_state_raises(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r3"
        with pytest.raises(PipelineViewError, match=r"^Bad release state"):
            ReleaseView(mock_pipeline_config(), this_context)

    def test_deployments_clean(self, mock_pipeline_config):
        under_test = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)

        result = under_test.deployments

        result_tuples = {str(x.deployment_tuple) for x in result}
        assert result_tuples == {
            "sys1-c1-r1",
        }

    def test_deployments_empty(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        under_test = ReleaseView(mock_pipeline_config(), this_context)

        result = under_test.deployments

        result_tuples = {str(x.deployment_tuple) for x in result}
        assert not result_tuples

    def test_flatten_clean(self, mock_pipeline_config):
        under_test = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)

        result = under_test.flatten(MOCK_TO)

        assert len(result) == 2
        assert {x.context.client for x in result} == {
            "c1",
        }
        assert {x.context.system for x in result} == {
            "sys1",
        }
        assert {x.data.location_primary for x in result} == {
            "l1",
            "l2",
        }
        assert [x.data.to for x in result] == [
            MOCK_TO,
            MOCK_TO,
        ]



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from ruamel.yaml import YAML

from foodx_devops_tools.puff.arm import _linearize_parameters


class TestLinearizeParameters:
    def test_name(self):
        mock_base = {
            "name": "some-name",
            "p1": "bp1",
            "k2": "bk2",
            "environments": {
                "e1": dict(),
            },
        }
        expected_result = {
            "some-name.e1": {
                "environment": "e1",
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        result = _linearize_parameters(mock_base, "this.stub")

        assert result == expected_result

    def test_default_name(self):
        mock_base = {
            "name": "file-name",
            "default": {"name": "name-variable-value"},
            "p1": "bp1",
            "k2": "bk2",
            "environments": {
                "e1": dict(),
            },
        }
        expected_result = {
            "file-name.e1": {
                "environment": "e1",
                "name": "name-variable-value",
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        result = _linearize_parameters(mock_base, "this.stub")

        assert result == expected_result

    def test_empty_environment(self):
        mock_base = {
            "p1": "bp1",
            "k2": "bk2",
            "environments": {
                "e1": dict(),
            },
        }
        expected_result = {
            "this.stub.e1": {
                "environment": "e1",
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        result = _linearize_parameters(mock_base, "this.stub")

        assert result == expected_result

    def test_environment_empty_region(self):
        mock_base = {
            "p1": "bp1",
            "k2": "bk2",
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "regions": [
                        {
                            "r1": dict(),
                        }
                    ],
                },
            },
        }
        expected_result = {
            "this.stub.e1.r1": {
                "environment": "e1",
                "region": "r1",
                "p1": "e1p1",
                "k2": "bk2",
            },
        }
        result = _linearize_parameters(mock_base, "this.stub")

        assert result == expected_result

    def test_environment_none_region(self):
        mock_base = {
            "p1": "bp1",
            "k2": "bk2",
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "regions": [
                        {
                            "r1": None,
                        }
                    ],
                },
            },
        }
        expected_result = {
            "this.stub.e1.r1": {
                "environment": "e1",
                "region": "r1",
                "p1": "e1p1",
                "k2": "bk2",
            },
        }
        result = _linearize_parameters(mock_base, "this.stub")

        assert result == expected_result

    def test_environment_empty_region_empty_base(self):
        mock_base = {
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "regions": [
                        {
                            "r1": dict(),
                        }
                    ],
                },
            },
        }
        expected_result = {
            "this.stub.e1.r1": {
                "environment": "e1",
                "region": "r1",
                "p1": "e1p1",
            },
        }
        result = _linearize_parameters(mock_base, "this.stub")

        assert result == expected_result

    def test_environment_none_region_empty_base(self):
        mock_base = {
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "regions": [
                        {
                            "r1": None,
                        }
                    ],
                },
            },
        }
        expected_result = {
            "this.stub.e1.r1": {
                "environment": "e1",
                "region": "r1",
                "p1": "e1p1",
            },
        }
        result = _linearize_parameters(mock_base, "this.stub")

        assert result == expected_result

    def test_from_yaml(self):
        mock_data = """services:
  s1:
    p1: s1p1
    environments:
      e1:
        p1: s1e1p1
        regions:
          - r1:
              p1: s1e1r1p1
              p2:
                reference: 
                  keyVault:
                    id: idv
                  secretName: sn
        """
        expected_result = {
            "this.stub.s1.e1.r1": {
                "service": "s1",
                "environment": "e1",
                "region": "r1",
                "p1": "s1e1r1p1",
                "p2": {
                    "reference": {
                        "keyVault": {"id": "idv"},
                        "secretName": "sn",
                    },
                },
            },
        }
        yaml = YAML(typ="safe")
        yaml_data = yaml.load(mock_data)

        result = _linearize_parameters(yaml_data, "this.stub")

        assert result == expected_result

    def test_complex1_clean(self):
        mock_base = {
            "p1": "bp1",
            "bp2": "bp2",
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "e1p2": "e1p2",
                    "regions": [
                        {
                            "r1": dict(),
                        }
                    ],
                },
            },
            "services": {
                "s1": {
                    "s1p2": "s1p2",
                    "environments": {
                        "e1": {
                            "s1e1p2": "s1e1p2",
                            "regions": [
                                {
                                    "r1": {
                                        "p1": "s1e1r1p1",
                                    },
                                }
                            ],
                        },
                    },
                },
            },
        }
        expected_result = {
            "filestub.s1.e1.r1": {
                "service": "s1",
                "environment": "e1",
                "region": "r1",
                "p1": "s1e1r1p1",
                "bp2": "bp2",
                "e1p2": "e1p2",
                "s1p2": "s1p2",
                "s1e1p2": "s1e1p2",
            },
        }
        result = _linearize_parameters(mock_base, "filestub")

        assert result == expected_result

    def test_complex2_clean(self):
        mock_base = {
            "p1": "bp1",
            "bp2": "bp2",
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "e1p2": "e1p2",
                    "regions": [
                        {
                            "r1": dict(),
                        }
                    ],
                },
            },
            "services": {
                "s1": {
                    "s1p2": "s1p2",
                    "environments": {
                        "e1": {
                            "s1e1p2": "s1e1p2",
                            "regions": [
                                {
                                    "r1": {
                                        "p1": "s1e1r1p1",
                                    },
                                }
                            ],
                        },
                    },
                },
                "s2": {
                    "s2p2": "s2p2",
                    "environments": {
                        "e1": {
                            "s2e1p2": "s2e1p2",
                            "regions": [
                                {
                                    "r1": {
                                        "p1": "s2e1r1p1",
                                    },
                                }
                            ],
                        },
                    },
                },
            },
        }
        expected_result = {
            "filestub.s1.e1.r1": {
                "service": "s1",
                "environment": "e1",
                "region": "r1",
                "p1": "s1e1r1p1",
                "bp2": "bp2",
                "e1p2": "e1p2",
                "s1p2": "s1p2",
                "s1e1p2": "s1e1p2",
            },
            "filestub.s2.e1.r1": {
                "service": "s2",
                "environment": "e1",
                "region": "r1",
                "p1": "s2e1r1p1",
                "bp2": "bp2",
                "e1p2": "e1p2",
                "s2p2": "s2p2",
                "s2e1p2": "s2e1p2",
            },
        }
        result = _linearize_parameters(mock_base, "filestub")

        assert result == expected_result

    def test_complex3_clean(self):
        mock_base = {
            "p1": "bp1",
            "bp2": "bp2",
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "e1p2": "e1p2",
                    "regions": [
                        {
                            "r1": {
                                "e1r1p2": "e1r1p2",
                            },
                        }
                    ],
                },
            },
            "services": {
                "s1": {
                    "s1p2": "s1p2",
                    "environments": {
                        "e1": {
                            "s1e1p2": "s1e1p2",
                            "regions": [
                                {
                                    "r1": {
                                        "p1": "s1e1r1p1",
                                    },
                                }
                            ],
                        },
                    },
                },
                "s2": {
                    "s2p2": "s2p2",
                    "environments": {
                        "e1": {
                            "s2e1p2": "s2e1p2",
                            "regions": [
                                {
                                    "r1": {
                                        "p1": "s2e1r1p1",
                                    },
                                }
                            ],
                        },
                    },
                },
            },
        }
        expected_result = {
            "filestub.s1.e1.r1": {
                "service": "s1",
                "environment": "e1",
                "region": "r1",
                "p1": "s1e1r1p1",
                "bp2": "bp2",
                "e1r1p2": "e1r1p2",
                "e1p2": "e1p2",
                "s1p2": "s1p2",
                "s1e1p2": "s1e1p2",
            },
            "filestub.s2.e1.r1": {
                "service": "s2",
                "environment": "e1",
                "region": "r1",
                "p1": "s2e1r1p1",
                "bp2": "bp2",
                "e1r1p2": "e1r1p2",
                "e1p2": "e1p2",
                "s2p2": "s2p2",
                "s2e1p2": "s2e1p2",
            },
        }
        result = _linearize_parameters(mock_base, "filestub")

        assert result == expected_result

    def test_complex4_clean(self):
        mock_base = {
            "p1": "bp1",
            "bp2": "bp2",
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "e1p2": "e1p2",
                    "regions": [
                        {
                            # None region specified.
                            "r1": None,
                        }
                    ],
                },
            },
            "services": {
                "s1": {
                    "s1p2": "s1p2",
                    "environments": {
                        "e1": {
                            "s1e1p2": "s1e1p2",
                            "regions": [
                                {
                                    "r1": {
                                        "p1": "s1e1r1p1",
                                    },
                                }
                            ],
                        },
                    },
                },
                "s2": {
                    "s2p2": "s2p2",
                    "environments": {
                        "e1": {
                            "s2e1p2": "s2e1p2",
                            "regions": [
                                {
                                    "r1": {
                                        "p1": "s2e1r1p1",
                                    },
                                }
                            ],
                        },
                    },
                },
            },
        }
        expected_result = {
            "filestub.s1.e1.r1": {
                "service": "s1",
                "environment": "e1",
                "region": "r1",
                "p1": "s1e1r1p1",
                "bp2": "bp2",
                "e1p2": "e1p2",
                "s1p2": "s1p2",
                "s1e1p2": "s1e1p2",
            },
            "filestub.s2.e1.r1": {
                "service": "s2",
                "environment": "e1",
                "region": "r1",
                "p1": "s2e1r1p1",
                "bp2": "bp2",
                "e1p2": "e1p2",
                "s2p2": "s2p2",
                "s2e1p2": "s2e1p2",
            },
        }
        result = _linearize_parameters(mock_base, "filestub")

        assert result == expected_result

    def test_complex5_clean(self):
        mock_base = {
            "p1": "bp1",
            "bp2": "bp2",
            "environments": {
                "e1": {
                    "p1": "e1p1",
                    "e1p2": "e1p2",
                    "regions": [
                        {
                            "r1": {
                                "p1": "s1e1r1p1",
                            },
                        }
                    ],
                },
            },
            "services": {
                "s1": {
                    "s1p2": "s1p2",
                    "environments": {
                        "e1": {
                            "s1e1p2": "s1e1p2",
                            "regions": [
                                {
                                    "r1": None,
                                }
                            ],
                        },
                    },
                },
                "s2": {
                    "s2p2": "s2p2",
                    "environments": {
                        "e1": {
                            "s2e1p2": "s2e1p2",
                            "regions": [
                                {
                                    "r1": {
                                        "p1": "s2e1r1p1",
                                    },
                                }
                            ],
                        },
                    },
                },
            },
        }
        expected_result = {
            "filestub.s1.e1.r1": {
                "service": "s1",
                "environment": "e1",
                "region": "r1",
                "p1": "s1e1r1p1",
                "bp2": "bp2",
                "e1p2": "e1p2",
                "s1p2": "s1p2",
                "s1e1p2": "s1e1p2",
            },
            "filestub.s2.e1.r1": {
                "service": "s2",
                "environment": "e1",
                "region": "r1",
                "p1": "s2e1r1p1",
                "bp2": "bp2",
                "e1p2": "e1p2",
                "s2p2": "s2p2",
                "s2e1p2": "s2e1p2",
            },
        }
        result = _linearize_parameters(mock_base, "filestub")

        assert result == expected_result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import pathlib

import pytest

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.deploy_me._deployment import (
    AzureSubscriptionConfiguration,
    DeploymentState,
    DeploymentStatus,
    deploy_application,
)
from foodx_devops_tools.pipeline_config import (
    IterationContext,
    PipelineConfiguration,
)
from foodx_devops_tools.pipeline_config.puff_map import PuffMapGeneratedFiles
from tests.ci.support.pipeline_config import MOCK_RESULTS

MOCK_ITERATION_CONTEXT = IterationContext()
MOCK_ITERATION_CONTEXT.append("some")
MOCK_ITERATION_CONTEXT.append("context")

MOCK_CONTEXT = str(MOCK_ITERATION_CONTEXT)

mock_pipeline_config = PipelineConfiguration.parse_obj(MOCK_RESULTS)
MOCK_APPLICATION_DATA = mock_pipeline_config.frames.frames["f1"].applications[
    "a1"
]


@pytest.fixture()
def prep_data(mock_async_method, mock_flattened_deployment):
    app_data = copy.deepcopy(MOCK_APPLICATION_DATA)

    deployment_data = mock_flattened_deployment[0]
    deployment_data.data.iteration_context.append("a1")
    deployment_data.context.frame_name = "f1"
    deployment_data.context.application_name = "a1"
    deployment_data.context.release_state = "r1"
    deployment_data.context.azure_subscription_name = "sys1_c1_r1a"
    deployment_data.data.puff_map = PuffMapGeneratedFiles.parse_obj(
        {"puff_map": MOCK_RESULTS["puff_map"]}
    ).puff_map

    mock_deploy = mock_async_method(
        "foodx_devops_tools.deploy_me._deployment.deploy_resource_group"
    )
    mock_puff = mock_async_method(
        "foodx_devops_tools.utilities.templates.run_puff"
    )
    mock_async_method(
        "foodx_devops_tools.deploy_me._deployment.login_service_principal"
    )
    mock_async_method(
        "foodx_devops_tools.utilities.templates._prepare_working_directory"
    )
    mock_async_method("foodx_devops_tools.utilities.templates._apply_template")

    return mock_deploy, mock_puff, deployment_data, app_data


class DeploymentChecks:
    async def check_static_resource_group(
        self, enable_validation: bool, prep_data
    ):
        mock_deploy, mock_puff, deployment_data, app_data = prep_data

        this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=0.1)
        application_deployment_data = copy.deepcopy(deployment_data)
        application_deployment_data.data.frame_folder = pathlib.Path(
            "some/path"
        )
        await deploy_application(
            app_data,
            application_deployment_data,
            this_status,
            enable_validation,
        )

        mock_puff.assert_called_once()

        return mock_deploy


class TestValidation(DeploymentChecks):
    @pytest.mark.asyncio
    async def test_static_resource_group(
        self, default_override_parameters, mock_verify_puff_target, prep_data
    ):
        enable_validation = True

        mock_deploy = await self.check_static_resource_group(
            enable_validation, prep_data
        )

        expected_parameters = default_override_parameters(prep_data[2])
        mock_deploy.assert_called_once_with(
            "a1_group-123456",
            pathlib.Path("some/path/working/w/a1.json"),
            pathlib.Path("some/path/working/w/some/puff_map/jinjad.path"),
            "l1",
            "Incremental",
            AzureSubscriptionConfiguration(subscription_id="sys1_c1_r1a"),
            deployment_name="a1_a1l1_123456",
            override_parameters=expected_parameters,
            validate=True,
        )


class TestDeployment(DeploymentChecks):
    @pytest.mark.asyncio
    async def test_static_resource_group(
        self, default_override_parameters, mock_verify_puff_target, prep_data
    ):
        enable_validation = False

        mock_deploy = await self.check_static_resource_group(
            enable_validation, prep_data
        )

        expected_parameters = default_override_parameters(prep_data[2])
        mock_deploy.assert_called_once_with(
            "a1_group",
            pathlib.Path("some/path/working/w/a1.json"),
            pathlib.Path("some/path/working/w/some/puff_map/jinjad.path"),
            "l1",
            "Incremental",
            AzureSubscriptionConfiguration(subscription_id="sys1_c1_r1a"),
            deployment_name="a1_a1l1_123456",
            override_parameters=expected_parameters,
            validate=False,
        )

    @pytest.mark.asyncio
    async def test_application_skip(self, prep_data):
        enable_validation = False
        mock_deploy, mock_puff, deployment_data, app_data = prep_data

        updated = copy.deepcopy(app_data)
        updated.steps[0].resource_group = None
        this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=0.1)
        application_deployment_data = copy.deepcopy(deployment_data)
        application_deployment_data.data.frame_folder = pathlib.Path(
            "some/path"
        )
        application_deployment_data.data.to = StructuredTo(
            frame="f1", application="a2"
        )

        await deploy_application(
            updated,
            application_deployment_data,
            this_status,
            enable_validation,
        )

        status = await this_status.read("a1")
        assert status.code == DeploymentState.ResultType.skipped



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio

import pytest

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.deploy_me._deployment import (
    ApplicationDeploymentSteps,
    DeploymentState,
    DeploymentStatus,
    FlattenedDeployment,
    deploy_frame,
)
from foodx_devops_tools.deploy_me.exceptions import DeploymentTerminatedError
from foodx_devops_tools.pipeline_config import IterationContext

MOCK_ITERATION_CONTEXT = IterationContext()
MOCK_ITERATION_CONTEXT.append("some")
MOCK_ITERATION_CONTEXT.append("context")
MOCK_CONTEXT = str(MOCK_ITERATION_CONTEXT)


@pytest.fixture()
def mock_application_deploy(mock_async_method, prep_frame_data):
    deployment_data, frame_data = prep_frame_data
    mock_application = mock_async_method(
        "foodx_devops_tools.deploy_me._deployment.deploy_application"
    )

    return mock_application, deployment_data, frame_data


class MockAppDeploy:
    def __init__(self, status):
        self.status = status

    async def __call__(
        self,
        application_data: ApplicationDeploymentSteps,
        deployment_data: FlattenedDeployment,
        application_status: DeploymentStatus,
        enable_validation: bool,
    ) -> None:
        this_context = str(deployment_data.data.iteration_context)
        await application_status.initialize(this_context)
        await application_status.write(this_context, self.status)


@pytest.mark.asyncio
async def test_clean(
    mocker, mock_completion_event, pipeline_parameters, prep_frame_data
):
    cli_options = pipeline_parameters()
    deployment_data, frame_data = prep_frame_data
    mock_app_deploy = MockAppDeploy(DeploymentState.ResultType.success)

    mocker.patch(
        "foodx_devops_tools.deploy_me._deployment.deploy_application" "",
        mock_app_deploy,
    )

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )

    result = await this_status.read("f1")
    assert result.code == DeploymentState.ResultType.success


@pytest.mark.asyncio
async def test_app_failed(
    mocker, mock_completion_event, pipeline_parameters, prep_frame_data
):
    cli_options = pipeline_parameters()
    deployment_data, frame_data = prep_frame_data
    mock_app_deploy = MockAppDeploy(DeploymentState.ResultType.failed)

    mocker.patch(
        "foodx_devops_tools.deploy_me._deployment.deploy_application" "",
        mock_app_deploy,
    )

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )

    result = await this_status.read("f1")
    assert result.code == DeploymentState.ResultType.failed


@pytest.mark.asyncio
async def test_app_cancelled(
    mocker, mock_completion_event, pipeline_parameters, prep_frame_data
):
    cli_options = pipeline_parameters()
    deployment_data, frame_data = prep_frame_data
    mock_app_deploy = MockAppDeploy(DeploymentState.ResultType.cancelled)

    mocker.patch(
        "foodx_devops_tools.deploy_me._deployment.deploy_application" "",
        mock_app_deploy,
    )

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )

    result = await this_status.read("f1")
    assert result.code == DeploymentState.ResultType.cancelled


@pytest.mark.asyncio
async def test_correct_results(
    mocker, mock_completion_event, pipeline_parameters, prep_frame_data
):
    """Test that a frame is correctly reporting the results of it's
    applications (and not the results of other frames)."""
    cli_options = pipeline_parameters()
    deployment_data, frame_data = prep_frame_data
    mock_app_deploy = MockAppDeploy(DeploymentState.ResultType.success)

    mocker.patch(
        "foodx_devops_tools.deploy_me._deployment.deploy_application" "",
        mock_app_deploy,
    )

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)

    # ensure that another frame has failed to avoid masking a false positive
    # in frame status reporting if it is accidentally reporting
    await this_status.initialize("other-f")
    await this_status.write("other-f", DeploymentState.ResultType.failed)

    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )

    result = await this_status.read("f1")
    assert result.code == DeploymentState.ResultType.success


@pytest.mark.asyncio
async def test_dependency_failed(
    mock_application_deploy, mock_completion_event, pipeline_parameters
):
    cli_options = pipeline_parameters
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await this_status.initialize(dependency_frame)
    await this_status.write(dependency_frame, DeploymentState.ResultType.failed)

    await check_is_cancelled(
        this_status, frame_data, deployment_data, cli_options
    )


async def check_is_cancelled(
    this_status, frame_data, deployment_data, pipeline_parameters
):
    cli_options = pipeline_parameters()
    with pytest.raises(
        DeploymentTerminatedError, match=r"^cancelled due to dependency failure"
    ):
        await deploy_frame(
            frame_data,
            deployment_data,
            this_status,
            cli_options,
        )


@pytest.mark.asyncio
async def test_dependency_cancelled(
    mock_application_deploy, mock_completion_event, pipeline_parameters
):
    cli_options = pipeline_parameters
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await this_status.initialize(dependency_frame)
    await this_status.write(
        dependency_frame, DeploymentState.ResultType.cancelled
    )

    await check_is_cancelled(
        this_status, frame_data, deployment_data, cli_options
    )


@pytest.mark.asyncio
async def test_multiple_dependencies_one_failed(
    mock_application_deploy, mock_completion_event, pipeline_parameters
):
    cli_options = pipeline_parameters
    dependencies = {
        "df1": DeploymentState.ResultType.success,
        "df2": DeploymentState.ResultType.failed,
    }

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = list(dependencies.keys())

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    for this_df in dependencies.keys():
        await this_status.initialize(this_df)
        await this_status.write(this_df, dependencies[this_df])

    await check_is_cancelled(
        this_status, frame_data, deployment_data, cli_options
    )


@pytest.mark.asyncio
async def test_dependencies_success(
    mock_application_deploy, mock_completion_event, pipeline_parameters
):
    cli_options = pipeline_parameters()
    dependency_frame = "other-frame"
    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await this_status.initialize(dependency_frame)
    await this_status.write(
        dependency_frame, DeploymentState.ResultType.success
    )

    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )


async def delayed_completion(
    name: str,
    status: DeploymentStatus,
    duration_seconds: float,
    final_state: DeploymentState.ResultType,
):
    await status.write(name, DeploymentState.ResultType.in_progress)
    await asyncio.sleep(duration_seconds)
    await status.write(name, final_state)


@pytest.mark.asyncio
async def test_pausing(
    mock_application_deploy, mock_completion_event, pipeline_parameters
):
    cli_options = pipeline_parameters()
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await this_status.initialize(dependency_frame)

    await asyncio.create_task(
        delayed_completion(
            dependency_frame, this_status, 2, DeploymentState.ResultType.success
        )
    )

    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )


@pytest.mark.asyncio
async def test_in_progress_timeout(
    mock_application_deploy, mock_completion_event, pipeline_parameters
):
    cli_options = pipeline_parameters(
        monitor_sleep_seconds=2, wait_timeout_seconds=1
    )
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await this_status.initialize(dependency_frame)

    await asyncio.create_task(
        delayed_completion(
            dependency_frame,
            this_status,
            0.5,
            DeploymentState.ResultType.in_progress,
        )
    )

    with pytest.raises(
        DeploymentTerminatedError,
        match=r"^deployment cancelled due to timeout waiting for "
        "dependency completion",
    ):
        await deploy_frame(
            frame_data,
            deployment_data,
            this_status,
            cli_options,
        )


@pytest.mark.asyncio
async def test_frame_skipped(
    mock_application_deploy, mock_completion_event, pipeline_parameters
):
    cli_options = pipeline_parameters(
        monitor_sleep_seconds=2, wait_timeout_seconds=1
    )
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=1)
    await this_status.initialize(dependency_frame)

    await asyncio.create_task(
        delayed_completion(
            dependency_frame,
            this_status,
            0.5,
            DeploymentState.ResultType.in_progress,
        )
    )
    deployment_data.data.to = StructuredTo(frame="f2")

    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )

    f1_status = await this_status.read("f1")
    assert f1_status.code == DeploymentState.ResultType.skipped



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""I/O related utilities."""

import pathlib
import typing

import ruamel.yaml

from .ansible import managed_file_decrypt


def acquire_token(password_file: typing.IO) -> str:
    """
    Read a password token from the specified stream.

    Args:
        password_file: Input file stream to read password token from.

    Returns:
        Read token.
    """
    with password_file:
        content = password_file.read()

    return content


def load_encrypted_data(this_file: pathlib.Path, decrypt_token: str) -> dict:
    """Load YAML data from an Ansible Vault encrypted file."""
    yaml = ruamel.yaml.YAML(typ="safe")
    with managed_file_decrypt(this_file, decrypt_token) as decrypted_stream:
        yaml_data = yaml.load(decrypted_stream)

    return yaml_data



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""General utilities supporting foodx-devops-tools package."""

from .command import (  # noqa: F401
    CapturedStreams,
    CommandArgs,
    run_async_command,
    run_command,
)
from .git import get_sha  # noqa: F401
from .io import acquire_token  # noqa: F401



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest

from foodx_devops_tools.utilities.templates import (
    TemplateFiles,
    TemplatePaths,
    _apply_template,
    _construct_arm_template_parameter_paths,
    json_inlining,
    prepare_deployment_files,
)


@pytest.fixture()
def mock_run(mock_async_method, mocker):
    mock_puff = mock_async_method(
        "foodx_devops_tools.utilities.templates.run_puff"
    )
    mock_template = mock_async_method(
        "foodx_devops_tools.utilities.templates.FrameTemplates"
        ".apply_template"
    )
    mocker.patch("pathlib.Path.is_file", return_value=True)

    return mock_puff, mock_template


class TestPrepareArmTemplateParameterPaths:
    def test_clean(self):
        mock_data = TemplatePaths(
            source=pathlib.Path("source/puff/file.yml"),
            target=pathlib.Path("target/json/file.json"),
        )

        result = _construct_arm_template_parameter_paths(mock_data)

        assert (
            result == mock_data.target.parent / f"jinjad."
            f"{mock_data.target.name}"
        )


class TestPrepareDeploymentFiles:
    MOCK_PARAMETERS = {
        "k1": "v1",
        "k2": "v2",
    }

    @pytest.mark.asyncio
    async def test_clean(self, mock_run):
        mock_path = pathlib.Path("some/file")
        mock_templates = TemplateFiles(
            arm_template=TemplatePaths(
                source=mock_path,
                target=mock_path,
            ),
            arm_template_parameters=TemplatePaths(
                source=mock_path,
                target=mock_path,
            ),
        )
        mock_puff, _ = mock_run

        result = await prepare_deployment_files(
            mock_templates, self.MOCK_PARAMETERS
        )

        assert result.arm_template == mock_path
        assert result.parameters == pathlib.Path("some/jinjad.file")
        mock_puff.assert_called_once_with(
            mock_path,
            False,
            False,
            disable_ascii_art=True,
            output_dir=pathlib.Path("some"),
        )

    @pytest.mark.asyncio
    async def test_jinja2_puff(self, mock_run, mocker):
        mock_puff, mock_template = mock_run
        mock_templates = TemplateFiles(
            arm_template=TemplatePaths(
                source=pathlib.Path("some/source/file"),
                target=pathlib.Path("some/target/file"),
            ),
            arm_template_parameters=TemplatePaths(
                source=pathlib.Path("some/path.yml"),
                target=pathlib.Path("some/target/generated.json"),
            ),
        )

        result = await prepare_deployment_files(
            mock_templates,
            self.MOCK_PARAMETERS,
        )

        assert result.arm_template == mock_templates.arm_template.target
        assert result.parameters == pathlib.Path(
            "some/target/jinjad.generated.json"
        )
        mock_puff.assert_called_once_with(
            mock_templates.arm_template_parameters.source,
            False,
            False,
            disable_ascii_art=True,
            output_dir=pathlib.Path("some/target"),
        )
        mock_template.assert_has_calls(
            [
                mocker.call(
                    mock_templates.arm_template_parameters.target.name,
                    pathlib.Path("some/target/jinjad.generated.json"),
                    self.MOCK_PARAMETERS,
                ),
                mocker.call(
                    mock_templates.arm_template.source.name,
                    mock_templates.arm_template.target,
                    self.MOCK_PARAMETERS,
                ),
            ],
            any_order=True,
        )

    @pytest.mark.asyncio
    async def test_jinja2_arm(self, mock_run, mocker):
        mock_puff, mock_template = mock_run
        mock_templates = TemplateFiles(
            arm_template=TemplatePaths(
                source=pathlib.Path("some/jinja2.path"),
                target=pathlib.Path("some/target/path"),
            ),
            arm_template_parameters=TemplatePaths(
                source=pathlib.Path("some/path.yml"),
                target=pathlib.Path("some/target/generated.json"),
            ),
        )

        result = await prepare_deployment_files(
            mock_templates,
            self.MOCK_PARAMETERS,
        )

        assert result.arm_template == mock_templates.arm_template.target
        assert result.parameters == pathlib.Path(
            "some/target/jinjad.generated.json"
        )
        mock_puff.assert_called_once_with(
            mock_templates.arm_template_parameters.source,
            False,
            False,
            disable_ascii_art=True,
            output_dir=pathlib.Path("some/target"),
        )
        mock_template.assert_has_calls(
            [
                mocker.call(
                    mock_templates.arm_template_parameters.target.name,
                    pathlib.Path("some/target/jinjad.generated.json"),
                    self.MOCK_PARAMETERS,
                ),
                mocker.call(
                    mock_templates.arm_template.source.name,
                    mock_templates.arm_template.target,
                    self.MOCK_PARAMETERS,
                ),
            ],
            any_order=True,
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Run resource group tests against an actual Azure Cloud subscription."""

import json
import logging
import uuid

import pytest

from foodx_devops_tools.azure.cloud.resource_group import (
    check_exists as check_resource_group_exists,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    validate as validate_resource_group,
)

from .support import unique_group_name  # noqa: F401

log = logging.getLogger(__name__)


class TestValidateResourceGroup:
    async def do_check(self, group_name: str, subscription):
        assert not await check_resource_group_exists(group_name, subscription)

    @pytest.mark.asyncio
    async def test_clean(
        self,
        unique_group_name,
        user_subscription,
        template_file,
        parameters_file,
    ):
        expected_group = unique_group_name
        expected_resource = str(uuid.uuid4()).replace("-", "")[0:20]
        expected_location = "eastus"

        with parameters_file.open(mode="r") as f:
            data = json.load(f)
        data["parameters"]["storageAccountName"]["value"] = expected_resource
        if isinstance(data["parameters"]["location"], str):
            data["parameters"]["location"] = {"value": expected_location}
        else:
            data["parameters"]["location"]["value"] = expected_location
        with parameters_file.open(mode="w") as f:
            json.dump(data, f)

        await validate_resource_group(
            expected_group,
            template_file,
            parameters_file,
            expected_location,
            "Complete",
            user_subscription,
        )

        await self.do_check(expected_group, user_subscription)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

from foodx_devops_tools.pipeline_config import load_service_principals
from tests.ci.support.ansible import encrypted_file


class TestLoadServicePrincipals:
    mock_encrypted_path = pathlib.Path("some/file")
    mock_password_path = pathlib.Path("other/file")

    def test_load(self):
        file_text = """
---
service_principals:
  sub1_name:
    id: 12345-id
    name: principal-name1
    secret: verysecret
  sub2_name:
    id: 123456-id
    name: principal-name2
    secret: verysecret
"""
        decrypt_token = "somesecret"
        with encrypted_file(file_text, decrypt_token) as (encrypted_file_path):
            result = load_service_principals(encrypted_file_path, decrypt_token)

        assert len(result.service_principals) == 2
        assert result.service_principals["sub1_name"].id == "12345-id"
        assert result.service_principals["sub1_name"].name == "principal-name1"
        assert result.service_principals["sub1_name"].secret == "verysecret"

        assert result.service_principals["sub2_name"].id == "123456-id"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import copy
import logging
import re
import typing

import click

from foodx_devops_tools.azure.cloud.auth import (
    AzureAuthenticationError,
    login_service_principal,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    AzureSubscriptionConfiguration,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    deploy as deploy_resource_group,
)
from foodx_devops_tools.pipeline_config import (
    ApplicationDefinition,
    ApplicationDeploymentSteps,
    ApplicationStepDelay,
    FlattenedDeployment,
    PipelineConfiguration,
    SingularFrameDefinition,
)
from foodx_devops_tools.pipeline_config.frames import (
    ApplicationStepDeploymentDefinition,
)
from foodx_devops_tools.pipeline_config.puff_map import PuffMapPaths
from foodx_devops_tools.puff import PuffError
from foodx_devops_tools.utilities.templates import prepare_deployment_files

from ._dependency_monitor import wait_for_dependencies
from ._exceptions import DeploymentError
from ._state import PipelineCliOptions
from ._status import DeploymentState, DeploymentStatus, all_success

log = logging.getLogger(__name__)

STATUS_SLEEP_SECONDS = 1


def any_failed(values: typing.List[DeploymentState]) -> bool:
    """Evaluate if any deployment states failed."""
    result = any([x.code == DeploymentState.ResultType.failed for x in values])
    return result


def all_cancelled(values: typing.List[DeploymentState]) -> bool:
    """Evaluate if all deployments were cancelled."""
    result = all(
        [x.code == DeploymentState.ResultType.cancelled for x in values]
    )
    return result


async def assess_results(
    results: typing.List[DeploymentState],
) -> DeploymentState:
    """
    Condense an array of deployment results into a single result.

    Args:
        results: Array of results to condense.

    Returns:
        Success, if all results are success. Failed otherwise.
    """
    if all_success(results):
        log.debug("assessed result success")
        this_result = DeploymentState(code=DeploymentState.ResultType.success)
    elif any_failed(results):
        log.debug("assessed result failed")
        this_result = DeploymentState(code=DeploymentState.ResultType.failed)
    elif all_cancelled(results):
        log.debug("assessed result cancelled")
        this_result = DeploymentState(code=DeploymentState.ResultType.cancelled)
    else:
        # fallback is failure.
        log.debug("fallback to failed")
        messages = [
            x.message
            for x in results
            if x.message and (x.code != DeploymentState.ResultType.success)
        ]
        this_result = DeploymentState(
            code=DeploymentState.ResultType.failed, message=str(messages)
        )

    return this_result


def _mangle_validation_resource_group(current_name: str, suffix: str) -> str:
    this_suffix = re.sub(r"[_.+]", "-", suffix)
    mangled_name = f"{current_name}-{this_suffix}"

    return mangled_name


def _make_secrets_object(key_values: dict) -> typing.List[dict]:
    """Construct secrets into object form required by Foodx ARM template."""
    result = list()
    for k, v in key_values.items():
        this_entry = {
            "enabled": True,
            "key": k,
            "value": v,
        }
        result.append(this_entry)

    return result


def _construct_override_parameters(
    deployment_data: FlattenedDeployment,
    static_secrets_enabled: typing.Optional[bool],
    step_context: str,
) -> typing.Dict[str, typing.Any]:
    result: typing.Dict[str, typing.Any] = {
        "locations": {
            "primary": deployment_data.data.location_primary,
            "secondary": deployment_data.data.location_secondary,
        },
        "tags": deployment_data.context.as_dict(),
    }
    if static_secrets_enabled:
        if deployment_data.data.static_secrets:
            # pass static secrets as a single object containing all the
            # secret key-value pairs.
            parameter_object: typing.Dict[str, typing.Any] = {
                "staticSecrets": _make_secrets_object(
                    deployment_data.data.static_secrets
                )
            }
            result.update(parameter_object)
        else:
            log.warning(
                "There are no static_secrets even though secrets have been"
                " enabled, {0}".format(step_context)
            )

    return result


async def _do_step_deployment(
    this_step: ApplicationStepDeploymentDefinition,
    deployment_data: FlattenedDeployment,
    puff_parameter_paths: PuffMapPaths,
    this_context: str,
    enable_validation: bool,
) -> None:
    step_context = f"{this_context}.{this_step.name}"

    log.debug(
        f"deployment_data.context, "
        f"{step_context}, {str(deployment_data.context)}"
    )
    log.debug(
        f"deployment_data.data, {step_context}, {str(deployment_data.data)}"
    )
    try:
        resource_group = this_step.resource_group

        await login_service_principal(deployment_data.data.azure_credentials)
        if enable_validation:
            log.info(f"validation deployment enabled, {step_context}")
            resource_group = _mangle_validation_resource_group(
                resource_group,
                deployment_data.context.pipeline_id,
            )
        else:
            log.info(f"deployment enabled, {step_context}")

        template_parameters = deployment_data.construct_template_parameters()
        log.debug(f"template parameters, {step_context}, {template_parameters}")

        template_files = deployment_data.construct_deployment_paths(
            this_step.arm_file,
            this_step.puff_file,
            puff_parameter_paths[this_step.name],
        )
        log.debug(f"template files, {template_files}")

        deployment_files = await prepare_deployment_files(
            template_files,
            template_parameters,
        )

        override_parameters = _construct_override_parameters(
            deployment_data, this_step.static_secrets, step_context
        )
        deployment_name = deployment_data.construct_deployment_name(
            this_step.name
        )
        this_subscription = AzureSubscriptionConfiguration(
            subscription_id=deployment_data.context.azure_subscription_name
        )
        await deploy_resource_group(
            resource_group,
            deployment_files.arm_template,
            deployment_files.parameters,
            deployment_data.data.location_primary,
            this_step.mode.value,
            this_subscription,
            deployment_name=deployment_name,
            override_parameters=override_parameters,
            validate=enable_validation,
        )
    except Exception as e:
        message = f"step deployment failed, {step_context}, {str(e)}"
        log.error(message)
        click.echo(click.style(message, fg="red"), err=True)
        raise


async def _deploy_step(
    this_step: ApplicationStepDeploymentDefinition,
    deployment_data: FlattenedDeployment,
    puff_parameter_data: PuffMapPaths,
    this_context: str,
    enable_validation: bool,
) -> None:
    step_context = "{0}.{1}".format(this_context, this_step.name)
    deploy_to = deployment_data.data.to
    if deploy_to.step and (this_step.name != deploy_to.step):
        log.warning(
            "application step skipped using deployment specifier, "
            "{0} skipped {1}".format(str(deploy_to), step_context)
        )
    else:
        await _do_step_deployment(
            this_step,
            deployment_data,
            puff_parameter_data,
            this_context,
            enable_validation,
        )
        log.info("application step succeeded, {0}".format(step_context))


async def _do_application_deployment(
    this_context: str,
    application_data: ApplicationDeploymentSteps,
    deployment_data: FlattenedDeployment,
    application_status: DeploymentStatus,
    enable_validation: bool,
) -> None:
    try:
        puff_frame_data = deployment_data.data.puff_map.frames[
            deployment_data.context.frame_name
        ]
        puff_application_data = puff_frame_data.applications[
            deployment_data.context.application_name
        ]
        puff_parameter_data = puff_application_data.arm_parameters_files[
            deployment_data.context.release_state
        ][deployment_data.context.azure_subscription_name]

        for this_step in application_data:
            if isinstance(this_step, ApplicationStepDeploymentDefinition):
                await _deploy_step(
                    this_step,
                    deployment_data,
                    puff_parameter_data,
                    this_context,
                    enable_validation,
                )
            elif isinstance(this_step, ApplicationStepDelay):
                message = "application steps paused for, {0} (seconds)".format(
                    this_step.delay_seconds
                )
                log.info(message)
                click.echo(message)
                await asyncio.sleep(this_step.delay_seconds)
            else:
                raise DeploymentError(
                    "Bad application step definition, "
                    "{0}".format(this_context)
                )

        log.info("application deployment succeeded, {0}".format(this_context))
        await application_status.write(
            this_context, DeploymentState.ResultType.success
        )
    except asyncio.CancelledError:
        message = "async cancelled signal"
        log.error(message)
        await application_status.write(
            this_context,
            DeploymentState.ResultType.cancelled,
            message,
        )
        raise
    except (AzureAuthenticationError, PuffError) as e:
        message = (
            "application deployment authentication "
            "failure, {0}, {1}, {2}".format(this_context, type(e), str(e))
        )
        await application_status.write(
            this_context,
            DeploymentState.ResultType.failed,
            message,
        )
        log.error(message)


async def deploy_application(
    application_data: ApplicationDefinition,
    deployment_data: FlattenedDeployment,
    application_status: DeploymentStatus,
    enable_validation: bool,
) -> None:
    """
    Deploy the steps of a frame application.

    Application steps must be deployed in sequence (serially).
    """
    this_context = str(deployment_data.data.iteration_context)
    try:
        message = "starting application deployment, {0}".format(this_context)
        log.info(message)
        click.echo(message)
        await application_status.initialize(this_context)

        await wait_for_dependencies(
            deployment_data.data.iteration_context,
            application_data.depends_on
            if application_data.depends_on
            else list(),
            application_status,
        )

        await application_status.write(
            this_context, DeploymentState.ResultType.in_progress
        )

        deploy_to = deployment_data.data.to
        application_name = deployment_data.context.application_name
        if deploy_to.application and (
            application_name != deploy_to.application
        ):
            await application_status.write(
                this_context,
                DeploymentState.ResultType.skipped,
                message="deployment targeted application, {0}".format(
                    str(deploy_to)
                ),
            )
        else:
            await _do_application_deployment(
                this_context,
                application_data.steps,
                deployment_data,
                application_status,
                enable_validation,
            )
    except Exception as e:
        message = "application deployment failed, {0}, {1}, {2}".format(
            this_context, type(e), str(e)
        )
        await application_status.write(
            this_context,
            DeploymentState.ResultType.failed,
            message,
        )
        log.exception(message)

    message = "application deployment completed, {0}".format(this_context)
    log.info(message)
    click.echo(click.style(message, fg="yellow"))


async def _do_frame_deployment(
    this_context: str,
    pipeline_parameters: PipelineCliOptions,
    deployment_data: FlattenedDeployment,
    frame_data: SingularFrameDefinition,
    frame_status: DeploymentStatus,
) -> None:
    # application status will show as "pending" until deployment activates.
    application_status = DeploymentStatus(
        this_context, pipeline_parameters.wait_timeout_seconds
    )
    try:
        wait_task = asyncio.create_task(
            application_status.wait_for_all_completed()
        )
        application_status.start_monitor()

        await wait_for_dependencies(
            deployment_data.data.iteration_context,
            frame_data.depends_on if frame_data.depends_on else list(),
            frame_status,
        )

        frame_deployment = copy.deepcopy(deployment_data)
        frame_deployment.data.frame_folder = frame_data.folder

        await asyncio.gather(
            *[
                deploy_application(
                    application_data,
                    frame_deployment.copy_add_application(application_name),
                    application_status,
                    pipeline_parameters.enable_validation,
                )
                for application_name, application_data in frame_data.applications.items()  # noqa: E501
            ],
            return_exceptions=False,
        )

        await wait_task

        message = "frame deployment completed, {0}".format(this_context)
        log.info(message)
        click.echo(click.style(message, fg="yellow"))
    except asyncio.TimeoutError:
        message = "timeout waiting for application deployment, {0}".format(
            deployment_data.data.iteration_context
        )
        log.error(message)
        click.echo(message, err=True)
        raise

    # read the state of the completed _applications_
    results = [
        await application_status.read(x)
        for x in await application_status.names()
    ]
    condensed_result = await assess_results(results)
    await frame_status.write(
        this_context, condensed_result.code, condensed_result.message
    )
    message = "frame deployment {0}, {1}".format(
        condensed_result.code.name, this_context
    )
    log.info(message)
    click.echo(
        click.style(
            message, fg=DeploymentStatus.STATE_COLOURS[condensed_result.code]
        )
    )


async def deploy_frame(
    frame_data: SingularFrameDefinition,
    deployment_data: FlattenedDeployment,
    frame_status: DeploymentStatus,
    pipeline_parameters: PipelineCliOptions,
) -> None:
    """
    Deploy the applications of a frame.

    Frame applications are deployed concurrently (in parallel).

    Raises:
        DeploymentTerminatedError: If any dependencies fail preventing
                                  completion.
    """
    this_context = str(deployment_data.data.iteration_context)
    message = "starting frame deployment, {0}".format(this_context)
    log.info(message)
    click.echo(message)

    await frame_status.initialize(this_context)
    await frame_status.write(
        this_context, DeploymentState.ResultType.in_progress
    )
    deploy_to = deployment_data.data.to
    frame_name = deployment_data.context.frame_name
    if deploy_to.frame and (frame_name != deploy_to.frame):
        await frame_status.write(
            this_context,
            DeploymentState.ResultType.skipped,
            message="deployment targeted frame, {0}".format(str(deploy_to)),
        )
    else:
        await _do_frame_deployment(
            this_context,
            pipeline_parameters,
            deployment_data,
            frame_data,
            frame_status,
        )


async def do_deploy(
    configuration: PipelineConfiguration,
    deployment_data: FlattenedDeployment,
    pipeline_parameters: PipelineCliOptions,
) -> DeploymentState:
    """
    Deploy the frames in a flattened deployment configuration.

    Raises:
        asyncio.TimeoutError: If the maximum wait time in
                              ``pipeline_parameters`` is exceeded.
    """
    this_frames = configuration.frames
    deployment_data.data.iteration_context.append(
        deployment_data.data.deployment_tuple
    )
    deployment_data.data.puff_map = configuration.puff_map

    frame_deployment_status = DeploymentStatus(
        str(deployment_data.data.iteration_context),
        timeout_seconds=pipeline_parameters.wait_timeout_seconds,
    )
    try:
        wait_task = asyncio.create_task(
            frame_deployment_status.wait_for_all_completed()
        )
        frame_deployment_status.start_monitor()

        await asyncio.gather(
            *[
                deploy_frame(
                    frame_data,
                    deployment_data.copy_add_frame(frame_name),
                    frame_deployment_status,
                    pipeline_parameters,
                )
                for frame_name, frame_data in this_frames.frames.items()
            ],
            return_exceptions=False,
        )
        await wait_task
    except asyncio.TimeoutError:
        message = "timeout waiting for frame deployments, {0}".format(
            deployment_data.data.iteration_context
        )
        log.error(message)
        click.echo(message, err=True)
        raise

    results = [
        await frame_deployment_status.read(x)
        for x in await frame_deployment_status.names()
    ]
    condensed_result = await assess_results(results)

    return condensed_result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import pathlib

import pytest

from foodx_devops_tools.deploy_me._deployment import (
    ApplicationStepDeploymentDefinition,
    FlattenedDeployment,
    PipelineCliOptions,
)
from foodx_devops_tools.patterns import SubscriptionData
from foodx_devops_tools.pipeline_config.frames import DeploymentMode
from foodx_devops_tools.pipeline_config.views import (
    AzureCredentials,
    DeployDataView,
)
from tests.ci.support.pipeline_config import MOCK_CONTEXT


@pytest.fixture()
def mock_completion_event(mock_async_method):
    mock_async_method(
        "foodx_devops_tools.deploy_me._deployment"
        ".DeploymentStatus.wait_for_all_completed"
    )


@pytest.fixture()
def pipeline_parameters():
    def _apply(
        enable_validation=True,
        monitor_sleep_seconds=30,
        wait_timeout_seconds=10,
    ):
        cli_options = PipelineCliOptions(
            enable_validation=enable_validation,
            monitor_sleep_seconds=monitor_sleep_seconds,
            wait_timeout_seconds=wait_timeout_seconds,
        )

        return cli_options

    return _apply


@pytest.fixture()
def default_override_parameters():
    def _apply(context):
        subscription_data = SubscriptionData.from_subscription_name(
            context.context.azure_subscription_name
        )
        result = {
            "locations": {
                "primary": context.data.location_primary,
                "secondary": context.data.location_secondary,
            },
            "tags": {
                "application_name": context.context.application_name,
                "client": context.context.client,
                "commit_sha": context.context.commit_sha,
                "frame_name": context.context.frame_name,
                "pipeline_id": context.context.pipeline_id,
                "release_id": context.context.release_id,
                "release_state": context.context.release_state,
                "resource_suffix": subscription_data.resource_suffix,
                "subscription_name": context.context.azure_subscription_name,
                "system": context.context.system,
                "tenant_name": context.context.azure_tenant_name,
            },
        }

        return result

    return _apply


@pytest.fixture()
def mock_deploystep_context():
    mock_context = copy.deepcopy(MOCK_CONTEXT)
    mock_context.application_name = "app-name"
    mock_context.azure_tenant_name = "t1"
    mock_context.azure_subscription_name = "sys1_c1_r1a"
    mock_context.client = "c1"
    mock_context.system = "sys1"
    mock_context.frame_name = "f1"

    data_view = DeployDataView(
        AzureCredentials(
            userid="abc",
            secret="verysecret",
            subscription="sys1_c1_r1a",
            name="n",
            tenant="123abc",
        ),
        "a-b-c",
        "uswest2",
        "r1",
        "some.where",
        dict(),
        ["a", "p"],
        dict(),
    )
    data_view.frame_folder = pathlib.Path("frame/folder")

    arguments = {
        "this_step": ApplicationStepDeploymentDefinition(
            mode=DeploymentMode.incremental,
            name="this_step",
            arm_file=pathlib.Path("arm.file"),
            puff_file=pathlib.Path("puff.file"),
            resource_group="rgn",
        ),
        "deployment_data": FlattenedDeployment(
            context=mock_context, data=data_view
        ),
        "puff_parameter_data": {
            "this_step": pathlib.Path("puff_generated.path"),
        },
        "this_context": "some.context",
        "enable_validation": False,
    }
    return arguments



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest

from foodx_devops_tools.file_maintainer_entry import (
    RunMonitor,
    _clean_filesystem,
    _do_delete,
    _get_filesystem_capacity_used,
    _run_maintainer,
)

MOCK_PATH = pathlib.Path("some/path")
MOCK_THRESHOLD = 80


@pytest.fixture()
def mock_fs_actions(mock_async_method, mocker):
    def _apply(expected_capacity):
        mock_clean = mock_async_method(
            "foodx_devops_tools.file_maintainer_entry" "._clean_filesystem"
        )
        mocker.patch(
            "foodx_devops_tools.file_maintainer_entry"
            "._get_filesystem_capacity_used",
            return_value=expected_capacity,
        )

        return mock_clean

    return _apply


class TestRunMonitor:
    def test_keep_running_persistence(self):
        under_test = RunMonitor(10)
        # should always be true, but we cannot test "infinity"
        assert under_test.keep_running()
        assert under_test.keep_running()
        assert under_test.keep_running()

    def test_not_keep_running_no_persistence(self):
        under_test = RunMonitor(None)

        assert under_test.keep_running()
        assert not under_test.keep_running()
        assert not under_test.keep_running()

    @pytest.mark.asyncio
    async def test_sleep_persistence(self, mock_async_method):
        mock_sleep = mock_async_method(
            "foodx_devops_tools.file_maintainer_entry" ".asyncio.sleep"
        )
        under_test = RunMonitor(10)

        await under_test.sleep()

        mock_sleep.assert_called_once_with(600)

    @pytest.mark.asyncio
    async def test_no_sleep_no_persistence(self, mock_async_method):
        mock_sleep = mock_async_method(
            "foodx_devops_tools.file_maintainer_entry" ".asyncio.sleep"
        )
        under_test = RunMonitor(None)

        await under_test.sleep()

        mock_sleep.assert_not_called()


class TestRunMaintainer:
    @pytest.mark.asyncio
    async def test_below_threshold_no_clean(self, mock_fs_actions):
        mock_clean = mock_fs_actions(MOCK_THRESHOLD - 1)

        await _run_maintainer(MOCK_PATH, None, MOCK_THRESHOLD)

        mock_clean.assert_not_called()

    @pytest.mark.asyncio
    async def test_equal_threshold_clean(self, mock_fs_actions):
        mock_clean = mock_fs_actions(MOCK_THRESHOLD)

        await _run_maintainer(MOCK_PATH, None, MOCK_THRESHOLD)

        mock_clean.assert_called_once()

    @pytest.mark.asyncio
    async def test_above_threshold_clean(self, mock_fs_actions):
        mock_clean = mock_fs_actions(MOCK_THRESHOLD + 1)

        await _run_maintainer(MOCK_PATH, None, MOCK_THRESHOLD)

        mock_clean.assert_called_once()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Declarations relating to deployment management."""

import dataclasses
import re
import typing

TUPLE_REGEX_PATTERN = (
    r"(?P<system>[a-z0-9_]+)-(?P<client>[a-z0-9_]+)-("
    r"?P<release_state>[a-z0-9_]+)"
)


class DeploymentTupleError(Exception):
    """Problem parsing deployment tuple text."""


T = typing.TypeVar("T", bound="DeploymentTuple")


@dataclasses.dataclass
class DeploymentTuple:
    """System deployment tuple."""

    client: str
    release_state: str
    system: str

    def __str__(self: T) -> str:
        """Convert ``DeploymentTuple` object to normalized representation."""
        result = "-".join([self.system, self.client, self.release_state])
        return result

    @classmethod
    def parse(cls: typing.Type[T], text: str) -> T:
        """
        Convert deployment tuple text to ``DeploymentTuple`` object.

        Args:
            text: Deployment tuple text to be parsed.

        Returns:
            ``DeploymentTuple`` object.
        """
        result = re.match(TUPLE_REGEX_PATTERN, text)
        if not result:
            raise DeploymentTupleError(
                "Error parsing tuple text, " "{0}".format(text)
            )

        client = result.group("client")
        system = result.group("system")
        release_state = result.group("release_state")

        return cls(client=client, release_state=release_state, system=system)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.patterns import SubscriptionData, SubscriptionNameError


def test_clean():
    name = "some_sub123scription_name"

    under_test = SubscriptionData.from_subscription_name(name)

    assert under_test.client == "sub123scription"
    assert under_test.system == "some"
    assert under_test.resource_suffix == "name"


EXPECTED_MESSAGE = r"^Subscription name does not conform to naming pattern"


def test_bad_raises1():
    name = "badname"

    with pytest.raises(SubscriptionNameError, match=EXPECTED_MESSAGE):
        SubscriptionData.from_subscription_name(name)


def test_bad_raises2():
    name = "not-valid-either"

    with pytest.raises(SubscriptionNameError, match=EXPECTED_MESSAGE):
        SubscriptionData.from_subscription_name(name)
